#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <elf.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>

char *sh_type_names[]={"SHT_NULL","SHT_PROGBITS","SHT_SYMTAB","SHT_STRTAB","SHT_RELA",
                       "SHT_HASH","SHT_DYNAMIC","SHT_NOTE","SHT_NOBITS","SHT_REL","SHT_SHLIB",
                       "SHT_DYNSYM"};

char *sh_flags_names[]={"","SHF_WRITE","SHF_ALLOC","SHF_WRITE, SHF_ALLOC","SHF_EXECINSTR",
                        "SHF_WRITE, SHF_EXECINSTR","SHF_ALLOC, SHF_EXECINSTR",
                        "SHF_WRITE, SHF_ALLOC, SHF_EXECINSTR"};

char *st_bind_names[]={"STB_LOCAL","STB_GLOBAL","STB_WEAK"};

char *st_type_names[]={"STT_NOTYPE","STT_OBJECT","STT_FUNC","STT_SECTION","STT_FILE"};

void main(int argc, char **argv)
{
   int fd, i;
   char *buf;
   struct stat st;
   Elf32_Ehdr *elfhdr;
   Elf32_Shdr *shdr;
   char **sections;

   if(stat(argv[1],&st)<0)
   {
      perror("stat");
      exit(-1);
   }
   buf=(char *)malloc(st.st_size);
   if(buf==NULL)
   {
      perror("malloc");
      exit(-1);
   }
   if((fd=open(argv[1],O_RDONLY))<0)
   {
      perror("open");
      exit(-1);
   }
   if(read(fd,buf,st.st_size)!=st.st_size)
   {
      perror("read");
      exit(-1);
   }
   close(fd);

   elfhdr=(Elf32_Ehdr *)buf;
   printf("Dumping ELF-Header:\n");
   printf("e_type = 0x%04x\n",elfhdr->e_type);
   printf("e_machine = 0x%04x\n",elfhdr->e_machine);
   printf("e_version = 0x%08x\n",elfhdr->e_version);
   printf("e_entry = 0x%08x\n",elfhdr->e_entry);
   printf("e_phoff = 0x%08x\n",elfhdr->e_phoff);
   printf("e_shoff = 0x%08x\n",elfhdr->e_shoff);
   printf("e_flags = 0x%08x\n",elfhdr->e_flags);
   printf("e_ehsize = %u\n",elfhdr->e_ehsize);
   printf("e_phentsize = %u\n",elfhdr->e_phentsize);
   printf("e_phnum = %u\n",elfhdr->e_phnum);
   printf("e_shentsize = %u\n",elfhdr->e_shentsize);
   printf("e_shnum = %u\n",elfhdr->e_shnum);
   printf("e_shstrndx = %u\n\n",elfhdr->e_shstrndx);

   shdr=(Elf32_Shdr *)(buf+elfhdr->e_shoff);
   // generate sections pointer array
   sections=(char **)malloc(elfhdr->e_shnum*sizeof(char *));
   if(sections==NULL)
   {
      perror("malloc");
      exit(-1);
   }
   for(i=0;i<elfhdr->e_shnum;i++)
   {
      sections[i]=(char *)buf+shdr[i].sh_offset;
   }

   for(i=0;i<elfhdr->e_shnum;i++)
   {
      printf("   Dumping section %d:\n",i);
      printf("   sh_name = 0x%08x (%s)\n",shdr[i].sh_name,sections[elfhdr->e_shstrndx]+shdr[i].sh_name);
      printf("   sh_type = 0x%08x (%s)\n",shdr[i].sh_type,sh_type_names[shdr[i].sh_type&0xF]);
      printf("   sh_flags = 0x%08x (%s)\n",shdr[i].sh_flags,sh_flags_names[shdr[i].sh_flags&0xF]);
      printf("   sh_addr = 0x%08x\n",shdr[i].sh_addr);
      printf("   sh_offset = 0x%08x\n",shdr[i].sh_offset);
      printf("   sh_size = 0x%08x\n",shdr[i].sh_size);
      printf("   sh_link = 0x%08x\n",shdr[i].sh_link);
      printf("   sh_info = 0x%08x\n",shdr[i].sh_info);
      printf("   sh_addralign = 0x%08x\n",shdr[i].sh_addralign);
      printf("   sh_entsize = 0x%08x\n",shdr[i].sh_entsize);
      printf("\n");
      if(shdr[i].sh_type==SHT_SYMTAB)
      {
         Elf32_Sym *sym;
         int j;

         sym=(Elf32_Sym *)(buf+shdr[i].sh_offset);
         printf("      Dumping symbol table:\n");
         for(j=0;j<shdr[i].sh_size/sizeof(Elf32_Sym);j++)
         {
            printf("      Symbol %d:\n",j);
            printf("      st_name = 0x%08x (%s)\n",sym[j].st_name,sections[shdr[i].sh_link]+sym[j].st_name);
            printf("      st_value = 0x%08x\n",sym[j].st_value);
            printf("      st_size = 0x%08x\n",sym[j].st_size);
            printf("      st_info = 0x%02x\n",sym[j].st_info);
            printf("         ELF32_ST_BIND = %s\n",st_bind_names[ELF32_ST_BIND(sym[j].st_info)]);
            printf("         ELF32_ST_TYPE = %s\n",st_type_names[ELF32_ST_TYPE(sym[j].st_info)]);
            printf("      st_other = 0x%02x\n",sym[j].st_other);
            printf("      st_shndx = %u\n",sym[j].st_shndx);
            printf("\n");
         }
      }
      else if(shdr[i].sh_type==SHT_REL)
      {
         Elf32_Rel *rel;
         int j;
         rel=(Elf32_Rel *)(buf+shdr[i].sh_offset);
         printf("      Dumping relocation table:\n");
         for(j=0;j<shdr[i].sh_size/sizeof(Elf32_Rel);j++)
         {
            printf("      r_offset = 0x%08x\n",rel[j].r_offset);
            printf("      r_info = 0x%08x\n",rel[j].r_info);
            printf("         symbol table index: %u\n",ELF32_R_SYM(rel[j].r_info));
            printf("         relocation type: %u\n",ELF32_R_TYPE(rel[j].r_info));
            printf("\n");
         }
      }
      else if(shdr[i].sh_type==SHT_RELA)
      {
         Elf32_Rela *rel;
         int j;
         rel=(Elf32_Rela *)(buf+shdr[i].sh_offset);
         printf("      Dumping relocation table:\n");
         for(j=0;j<shdr[i].sh_size/sizeof(Elf32_Rela);j++)
         {
            printf("      r_offset = 0x%08x\n",rel[j].r_offset);
            printf("      r_info = 0x%08x\n",rel[j].r_info);
            printf("         symbol table index: %u\n",ELF32_R_SYM(rel[j].r_info));
            printf("         relocation type: %u\n",ELF32_R_TYPE(rel[j].r_info));
            printf("      r_addend = 0x%08x\n",rel[j].r_addend);
            printf("\n");
         }
      }
   }

   free(sections);
   free(buf);
}
