/***********************************************************
 * kdisas.c
 *
 * disassembles a part of the kernel through reads on /dev/kmem
 *
 * (c) 2003 void.at
 * greuff <greuff@void.at>
 *
 ***********************************************************/

/* includes */
#include <stdio.h>
#include <netdb.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>
#include <errno.h>
#include <termios.h>
#include <sys/fcntl.h>
#include <asm/ioctls.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <arpa/inet.h>
#include <sys/ioctl.h>
#include <pty.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>

#include "libdis.h"

#define DEBUG

void readkmem(int kmemfd, unsigned long offset, void *buffer, unsigned int size);

int main(int argc, char **argv)
{
   int kmemfd;
   unsigned long addr;
   unsigned long len, pos, size;
   struct instr i;
   char *buf;

   printf("kdisas - (c) 2003 by void.at\ngreuff <greuff@void.at>\n\n");
   if(argc!=3)
   {
      help();
      return -1;
   }
   printf("Opening /dev/kmem for read/write...\n");
   kmemfd=open("/dev/kmem",O_RDONLY);
   if(kmemfd<0)
   {
      perror("Can't open /dev/kmem");
      return -1;
   }

   addr=strtoul(argv[1],NULL,16);
   len=strtoul(argv[2],NULL,10);

   printf("Using address 0x%08x length %ld\n",addr,len);
   buf=(char *)malloc(len);
   if(buf==NULL)
   {
      printf("can't allocate buf\n");
      exit(-1);
   }
   readkmem(kmemfd,addr,buf,len);
   pos=0;
   disassemble_init(0,INTEL_SYNTAX);
   while((pos+32)<len)
   {
      size=disassemble_address(buf+pos,&i);
      if(size)
      {
         printf("instruction (size=%d): %s %s,%s %s\n",size,i.mnemonic,
               i.dest,i.src,i.aux);
         pos+=size;
      }
      else
      {
         printf("unknown op\n");
         pos++;
      }
   }
   disassemble_cleanup();
   free(buf);
   close(kmemfd);
   return 0;  // to keep compiler happy
}

/***********************************************************************/

int help() {
   printf("Usage: kdisas 0xaddress length\n\n");
   return 0;
}

/***********************************************************************/

void readkmem(int kmemfd, unsigned long offset, void *buffer, unsigned int size)
{
   if(lseek(kmemfd,offset,SEEK_SET)!=offset)
   {
      perror("kmem lseek");
      exit(-1);
   }
   if(read(kmemfd,buffer,size)!=size)
   {
      perror("kmem read");
      exit(-1);
   }
}
