/* includes */
#include <stdio.h>
#include <netdb.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>
#include <errno.h>
#include <termios.h>
#include <sys/fcntl.h>
#include <asm/ioctls.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <arpa/inet.h>
#include <sys/ioctl.h>
#include <pty.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>
#include <linux/unistd.h>    // for syscall numbers (__NR_write)
#include "libdis.h"
#include "config.h"

// the interrupt descriptor table register
struct
{
   unsigned short limit;
   unsigned int base;
} __attribute__ ((packed)) idtr;

// an interrupt descriptor table entry
struct
{
   unsigned short off1;
   unsigned short sel;
   unsigned char none, flags;
   unsigned short off2;
} __attribute__ ((packed)) idt;

struct pathinfo
{
   unsigned short syscall;
   short path[32];
};

struct relocitem
{
   struct relocitem *next;
   char symbolname[32];
   unsigned long address;
   short pi_count;
   struct pathinfo pi[3];      // use a maximum of 3 different paths - so we can double-check the results
};

struct relocitem *relocinfo=NULL;
unsigned long sys_call_table;

void initrelocinfo();
void addrelocitem(char *symbolname, unsigned long address, short pi_count, struct pathinfo pi[3]);
struct relocitem *findsymbol(char *symbolname, int kmemfd);
unsigned long reloc_walk(int kmemfd, unsigned long startaddr, short *skips, struct relocitem *item);
unsigned long get_sys_call_table(int kmemfd);
void readkmem(int kmemfd, unsigned long offset, void *buffer, unsigned int size);

/* let the party start */
int main(int argc, char **argv)
{
   struct relocitem *item;
   int kmemfd;
   initrelocinfo();

   printf("Opening /dev/kmem for read/write...\n");
   kmemfd=open("/dev/kmem",O_RDWR);
   if(kmemfd<0)
   {
      perror("Can't open /dev/kmem");
      return -1;
   }

   printf("Determining the location of sys_call_table... ");
   sys_call_table=get_sys_call_table(kmemfd);
   if(sys_call_table==0)
   {
      close(kmemfd);
      return -1;
   }
   printf("0x%08x\n",sys_call_table);

   // get kernel symbols
   item=relocinfo;
   while(item)
   {
      struct relocitem *ri;
      ri=findsymbol(item->symbolname,kmemfd);
      printf("%08x X %s\n",item->address,item->symbolname);
      item=item->next;
   }

   close(kmemfd);
}


/******************************************************
 * initrelocinfo
 *
 * Initializes the paths to find kernel symbols.
 *
 ******************************************************/
void initrelocinfo()
{
   struct pathinfo pi[3];

   memset(pi,0,sizeof(pi));
   // BEWARE COMPLETELY BROKEN
  /* pi[0].syscall=__NR_setxattr;
   pi[0].path[0]=1;
   pi[0].path[1]=1;
   pi[0].path[2]=0;
   pi[0].path[3]=-1; */
   /* path over sys_readv: no inline functions in between, reliable */
   pi[1].syscall=__NR_readv;
   pi[1].path[0]=1;
   pi[1].path[1]=0;
   pi[1].path[2]=-1;
   /* path over sys_writev: no inline functions in between, reliable */
   pi[2].syscall=__NR_writev;
   pi[2].path[0]=1;
   pi[2].path[1]=0;
   pi[2].path[2]=-1;
   addrelocitem("kmalloc",0,3,pi);

   memset(pi,0,sizeof(pi));
   pi[0].syscall=__NR_reboot;
   pi[0].path[0]=1;
   pi[0].path[1]=-1;       /* kernel v2.6: 0, -1 (introduces device_shutdown()) */
   pi[1].syscall=__NR_reboot;
   pi[1].path[0]=4;
   pi[1].path[1]=-1;       /* kernel v2.6: 0, -1 (introduces device_shutdown()) */
   addrelocitem("printk",0,2,pi);

 /*  memset(pi,0,sizeof(pi));
   pi[0].syscall=__NR_read;
   pi[0].path[0]=0;
   pi[0].path[1]=-1;
   pi[1].syscall=__NR_write;
   pi[1].path[0]=0;
   pi[1].path[1]=-1;
   pi[2].syscall=__NR_readv;
   pi[2].path[0]=0;
   pi[2].path[1]=-1;
   addrelocitem("fget",0,3,pi);

   memset(pi,0,sizeof(pi));
   pi[0].syscall=__NR_readv;
   pi[0].path[0]=2;
   pi[0].path[1]=-1;
   addrelocitem("fput",0,1,pi);

   memset(pi,0,sizeof(pi));
   pi[0].syscall=__NR_write;
   pi[0].path[0]=5;
   pi[0].path[1]=-1;
   addrelocitem("dput",0,1,pi);

   memset(pi,0,sizeof(pi));
   pi[0].syscall=__NR_open;
   pi[0].path[0]=2;
   pi[0].path[1]=-1;
   addrelocitem("filp_open",0,1,pi);

   memset(pi,0,sizeof(pi));
   pi[0].syscall=__NR_close;
   pi[0].path[0]=0;
   pi[0].path[1]=-1;
   pi[1].syscall=__NR_exit;
   pi[1].path[0]=0;
   pi[1].path[1]=7;
   pi[1].path[2]=0;
   pi[1].path[3]=-1;
   addrelocitem("filp_close",0,2,pi);

   memset(pi,0,sizeof(pi));
   pi[0].syscall=__NR_execve;
   pi[0].path[0]=1;
   pi[0].path[1]=0;
   pi[0].path[2]=0;
   pi[0].path[3]=0;
   pi[0].path[4]=-1;
   addrelocitem("path_init",0,1,pi);

   memset(pi,0,sizeof(pi));
   pi[0].syscall=__NR_execve;
   pi[0].path[0]=1;
   pi[0].path[1]=0;
   pi[0].path[2]=0;
   pi[0].path[3]=1;
   pi[0].path[4]=-1;
   addrelocitem("path_walk",0,1,pi);*/
}

/******************************************************
 * addrelocitem
 *
 * Adds a relocation item to the global list.
 *
 * \param symbolname Name of the item
 * \param address Address of the item - probably 0 because we don't know yet
 * \param pi_count number of paths
 * \param pi the paths to the symbol
 *
 ******************************************************/
void addrelocitem(char *symbolname, unsigned long address, short pi_count, struct pathinfo pi[3])
{
   int i;
   struct relocitem *item=(struct relocitem *)malloc(sizeof(struct relocitem));
   struct relocitem *p;
   if(item==NULL)
   {
      printf("out of memory\n");
      exit(-1);
   }
   item->next=NULL;
   item->pi_count=pi_count;
   strcpy(item->symbolname,symbolname);
   item->address=address;
   for(i=0;i<pi_count;i++)
   {
      item->pi[i].syscall=pi[i].syscall;
      memcpy(item->pi[i].path,pi[i].path,sizeof(pi[i].path));
   }

   p=relocinfo;
   if(p==NULL)
   {
      relocinfo=item;
   }
   else
   {
      while(p->next)
         p=p->next;
      p->next=item;
   }
}

/******************************************************
 * findsymbol
 *
 * tries to get the address of a symbol. The symbol has to have an entry in relocinfo
 * or else the function will abort with error.
 *
 * \param symbolname Name of the symbol we search for
 * \param kmemfd file descriptor to opened /dev/kmem
 * \param sys_call_table address of sys_call_table
 * \returns pointer to the item if found, NULL otherwise
 *
 ******************************************************/
struct relocitem *findsymbol(char *symbolname, int kmemfd)
{
   struct relocitem *item=relocinfo;
   int match=0, i;
   unsigned long addr;
   unsigned long retdaddr[3];

   if(item==NULL)
   {
      printf("item is NULL - exiting\n");
      exit(-1);
   }
   do
   {
      if(!strcmp(symbolname,item->symbolname))
         match=1;
      else
         item=item->next;
   } while(!match && item);
   if(!match)
   {
      printf("|%s| not found in relocinfo - exiting\n",symbolname);
      exit(-1);
   }

   /** ok do the hard stuff - find the entry point of the first function and
       walk down the path **/

   // initialize disassembler
   disassemble_init(0,INTEL_SYNTAX);

   // walk all available paths
   for(i=0;i<item->pi_count;i++)
   {
      readkmem(kmemfd,sys_call_table+4*item->pi[i].syscall,&addr,4);
      #ifdef DEBUG
      //printf("syscall %d location: 0x%08x\n",item->pi[i].syscall,addr);
      #endif
      //printf(" (syscall addr: 0x%08x) ",addr);
      retdaddr[i]=reloc_walk(kmemfd,addr,&(item->pi[i].path[0]),item);
   }

   // check for equality
   if(item->pi_count==3)
   {
      if(retdaddr[0]==retdaddr[1] && retdaddr[1]==retdaddr[2])
      {
         // yeah...
         item->address=retdaddr[0];
      }
      else
      {
         if(retdaddr[0]==retdaddr[1] || retdaddr[0]==retdaddr[2] || retdaddr[1]==retdaddr[2])
         {
            printf("unsharp match - 0x%08x 0x%08x 0x%08x\n",retdaddr[0],retdaddr[1],retdaddr[2]);
            if(retdaddr[0]==retdaddr[1]) item->address=retdaddr[0];
            if(retdaddr[0]==retdaddr[2]) item->address=retdaddr[0];
            if(retdaddr[1]==retdaddr[2]) item->address=retdaddr[1];
         }
         else
         {
            printf("Paths completely differ - 0x%08x 0x%08x 0x%08x\n",retdaddr[0],retdaddr[1],retdaddr[2]);
         }
      }
   }
   else if(item->pi_count==2)
   {
      if(retdaddr[0]==retdaddr[1])
      {
         item->address=retdaddr[0];
      }
      else
      {
         printf("Unequal paths - 0x%08x != 0x%08x\n",retdaddr[0],retdaddr[1]);
      }
   }
   else
   {
      // well, not much left to check - take what we have
      item->address=retdaddr[0];
   }

   disassemble_cleanup();

   return item;
}

/******************************************************
 * reloc_walk
 *
 * recursive function, processes the pathinfo pointed to by skips
 *
 * \param kmemfd file descriptor to opened /dev/kmem
 * \param startaddr address to start the search from
 * \param skips how many calls shall be skipped before entering a new recursion
 * \param item pointer to the item being processed
 * \returns address of the symbol
 *
 ******************************************************/
unsigned long reloc_walk(int kmemfd, unsigned long startaddr, short *skips, struct relocitem *item)
{
   struct instr in;
   unsigned char buffer[4096];    // should be enough?
   unsigned long addr=startaddr;
   signed long helpaddr;
   int i;
   int pos, size;
   int found;
   struct instr inbfr[5];   // ringbuffer with the 5 most recent instructions
   memset(inbfr,0,sizeof(inbfr));

   // skip *skips calls
   for(i=0;i<=*skips;i++)
   {
      readkmem(kmemfd,addr,buffer,sizeof(buffer));
      // find call
      pos=0;
      found=0;
      while((pos+32)<sizeof(buffer) && found==0)
      {
         size=disassemble_address(buffer+pos,&in);
         if(size)
         {
            if(!strcmp(in.mnemonic,"call"))
            {
               found=1;
            }
            else
            {
               pos+=size;
            }

            // copy instruction to ringbuffer and advance
            inbfr[0]=inbfr[1];
            inbfr[1]=inbfr[2];
            inbfr[2]=inbfr[3];
            inbfr[3]=inbfr[4];
            inbfr[4]=in;
         }
         else
         {
            printf("Choking! Can't disassemble at 0x%08x\n",startaddr+pos);
            pos++;
         }
      }
      if(found==0)
      {
         printf("Can't skip calls within the first %d bytes - no %d calls found\n",
               sizeof(buffer),*skips);
         return 0;
      }

      // pos "points" to the beginning of the call instruction - skip it and proceed
      // only skip if it's not the last-before-end run
      if(i!=*skips)
      {
         pos+=size;
      }
      addr+=pos;
   }

   // advance skips
   skips++;

   // addr now finally "points" to the call instruction we were seeking for
   // fine-grained checks follow
   if(*skips==-1)
   {
#ifdef DEBUG
 /*     printf("Dumping instruction ringbuffer\n");
      for(i=0;i<5;i++)
      {
         printf("Instruction %d: %s %s, %s [%s]\n",i,inbfr[i].mnemonic,inbfr[i].dest,inbfr[i].src,
            inbfr[i].aux);
      }*/
#endif
    /*  if(!strcmp(item->symbolname,"kmalloc"))
      {
         // There should be two pushs before the call. The first push
         // pushes GPF_KERNEL, which is 0x1F0 in in.dest
         if(!strcmp(inbfr[2].mnemonic,"push") && !strcmp(inbfr[2].dest,"0x1F0") &&
            !strcmp(inbfr[3].mnemonic,"push"))
         {
            printf("[:-)] ");
         }
         else if(!strcmp(inbfr[2].mnemonic,"mov") && strstr(inbfr[2].dest,"esp") &&
                 !strcmp(inbfr[2].src,"0x1F0") && !strcmp(inbfr[3].mnemonic,"mov") &&
                 strstr(inbfr[3].dest,"esp"))
         {
            printf("[:-)]");
         }
      } */
   }

   // - extract the address from it
#ifdef DEBUG
   //printf("i'm on it (size=%d): %s %s, %s [%s]\n",size,in.mnemonic,in.dest,in.src,in.aux);
#endif
   readkmem(kmemfd,addr+1,&helpaddr,4);
   addr+=helpaddr+size;
#ifdef DEBUG
   //printf("I believe the address is 0x%08x\n",addr);
#endif

   if(*skips!=-1)
   {
      return reloc_walk(kmemfd,addr,skips,item);
   }

   return addr;
}

/******************************************************
 * get_sys_call_table
 *
 * \param kmemfd file descriptor of opened /dev/kmem
 * \returns address of sys_call_table
 ******************************************************/
unsigned long get_sys_call_table(int kmemfd)
{
   unsigned long sys_call_off;
   unsigned char buffer[100];
   void *ptr;
   
   // read IDTR
   asm("sidt %0" : "=m" (idtr));

   // get IDT for int $0x80
   readkmem(kmemfd,idtr.base+0x80*8,&idt,sizeof(idt));

   // compute the beginning of the ISR of int 0x80
   sys_call_off=(idt.off2<<16)|idt.off1;

   // read the first 100 bytes of the ISR into memory to
   // locate the sys_call_table (see [1] for details)
   readkmem(kmemfd,sys_call_off,buffer,sizeof(buffer));

   ptr=memmem(buffer,sizeof(buffer),"\xff\x14\x85",3);

   if(ptr)
   {
      unsigned long sys_call_table=*(unsigned long *)(ptr+3);
      return sys_call_table;
   }
   else
   {
      printf("   unable to find sys_call_table!\n");
      return 0;
   }
}

/******************************************************
 * readkmem
 *
 * Helper function, reads bytes from /dev/kmem
 *
 * \param kmemfd file descriptor of opened /dev/kmem
 * \param offset offset in /dev/kmem ("address" in kernel space :-))
 * \param buffer buffer to write the data to
 * \param size amount of bytes to read
 ******************************************************/
void readkmem(int kmemfd, unsigned long offset, void *buffer, unsigned int size)
{
   if(lseek(kmemfd,offset,SEEK_SET)!=offset)
   {
      perror("kmem lseek");
      exit(-1);
   }
   if(read(kmemfd,buffer,size)!=size)
   {
      perror("kmem read");
      exit(-1);
   }
}
